#include <iostream>
#include <iterator>
#include <map>
#include "gomoryhu.hpp"
#include "mincut.hpp"
#include "genericGraph.hpp"
using namespace std;

//Constructor
GomoryHu::GomoryHu(Graph aGraph){
	myGraph = aGraph;
	set<int>* initSet = new set<int>();
	int i;
	for(i=1;i<=myGraph.numVertices;i++){
		initSet->insert(i);
	}
	currentSets.insert(initSet);
	treeEdges = (GenericEdge **)calloc(myGraph.numVertices-1, sizeof(GenericEdge *));
}

/**
 * Gomory Hu Algorithm
 */
void GomoryHu::solve(){
	int s=2, t=3, i;
	Graph *aGraph;
	
	for(i=0; i < myGraph.numVertices-1; i++){
		cout << "\nIt ----------------------------------" << i <<endl<<endl;
		set< set<int>* >::iterator pos;
	    for (pos = currentSets.begin(); pos != currentSets.end(); ++pos) {
	        if((*pos)->size()!=1){
	        	
	        	set<int>::iterator pos2 = (*pos)->begin();
	        	s=*(pos2);
	        	t=*(++pos2);
	        	
	        	printf("Select %d and %d \n",s, t);
	        	aGraph = graphReduction(*pos); 
	        	aGraph->printData();
	        	 // TODO. Now here we have to create a new graph by collapsing all child nodes to one node
	        	 // and computing the min cut in the resulting graph.	        	 
	        	Cut *myCut = mincut(aGraph->numVertices, aGraph->numEdges, aGraph->Adj, aGraph->capacities, aGraph->Edges, s, t);	        	
	        									
	        	printSet(*(myCut->S));
				printSet(*(myCut->Sbar));
				
				int k, rep;
				for(k=0;k<i;k++){ // Search edges so far in tree
					
					// If one of the ends of the edge is the initial set
					if(treeEdges[k]->myS1 == *pos){ 
						rep = setReps[treeEdges[k]->myS2];
						cout<< "REp= "<<rep+1 <<endl;
						printSet(*treeEdges[k]->myS2);
						if(myCut->S->find(rep) != myCut->S->end()){
							treeEdges[k]->myS1 = myCut->S;
						}else{
							treeEdges[k]->myS1 = myCut->Sbar;
						}
					}else if( treeEdges[k]->myS2 == *pos) {						
						rep = setReps[treeEdges[k]->myS1];
						cout<< "REp= "<<rep+1 <<endl;
						printSet(*treeEdges[k]->myS1);
						if(myCut->S->find(rep) != myCut->S->end()){
							treeEdges[k]->myS2 = myCut->S;
						}else{
							treeEdges[k]->myS2 = myCut->Sbar;
						}
					}
					
				}			
				
				
				filterSet(myCut->S, *pos);
				filterSet(myCut->Sbar, *pos);
	        	treeEdges[i] = new GenericEdge(myCut->S, myCut->Sbar, myCut->weight);
	        	
				printTree();
	        	currentSets.erase(pos);	        	
	        	currentSets.insert(myCut->S);
	        	currentSets.insert(myCut->Sbar);
	        	
    			printSets(currentSets);
	        	
	        	
	        	break;
	        }	        
	    }
	}
	
}


Graph* GomoryHu::graphReduction(set<int>* S){
	int i;
	setReps.clear();
	Graph* newGraph = new Graph(myGraph);
	set< set<int>* >::iterator myIt;
	for (myIt = currentSets.begin(),i=0; myIt != currentSets.end(); ++myIt) {
        if(*myIt != S)
        	setReps[*myIt] = collapseSet(*myIt,newGraph);
    }
    
    return newGraph; 
}


void GomoryHu::filterSet(set<int> *S, set<int> *fS){
	set<int>::iterator myIt;
	for(myIt=S->begin(); myIt != S->end(); myIt++){
		if(fS->find(*myIt) == fS->end()){
			S->erase(myIt);
		}
	}
}

int GomoryHu::collapseSet(set<int> *S, Graph *G){
	int rep, j;
	set<int>::iterator pos =  S->begin();
	rep = *(pos)-1;
	printSet(*S);
    for (pos++; pos != S->end(); pos++) {
        for(j=0;j< G->numEdges;j++){
        	G->Adj[rep][j] += G->Adj[*pos-1][j];
        	G->Adj[*pos-1][j] = 0;
        }
    }
    
    
    return rep+1;
}


/**
 * Prints a set of sets of integers
 */
void GomoryHu::printSets(set< set<int>* > someSets){
	cout << "The current sets are:" <<endl;
	// iterate over all elements and print them
    set< set<int>* >::iterator pos1;
    for (pos1 = someSets.begin(); pos1 != someSets.end(); ++pos1) {
        printSet(**pos1);
    }
}

/**
 * Prints a set of integers
 */
void GomoryHu::printSet(set<int> aSet){
	// iterate over all elements and print them
    set<int>::iterator pos;
    for (pos = aSet.begin(); pos != aSet.end(); ++pos) {
        cout << *pos << ' ';
    }
    cout << endl;
}

void GomoryHu::printTree(){
	int i; 
	cout << "EDGES\n";
	
	for(i=0;i<myGraph.numVertices-1 && treeEdges[i]!=NULL ;i++){
		
		cout<< "Edge "<<i<<endl;
		printSet(*treeEdges[i]->myS1);
		printSet(*treeEdges[i]->myS2);
		printf("%d\n",treeEdges[i]->myW);
	}
	
	
}
